;;;; wmiirc.scm
;
;; A library for writing wmii configuration scripts
;;
;; This wmiirc is for wmii 3.6
;;
; Copyright (c) 2008, Peter Bex
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Please report bugs, suggestions and ideas to the Chicken Trac
; ticket tracking system (assign tickets to user 'sjamaan'):
; http://trac.callcc.org

(declare
 (export wmii:connect wmii:write wmii:read
	 wmii:read-lines wmii:directory
	 wmii:colrules wmii:colrules-set!
	 wmii:tagrules wmii:tagrules-set!
	 wmii:key-code->string wmii:string->key-code
	 wmii:grabbed-keys wmii:grabbed-keys-set!
	 wmii:event-handlers wmii:event-handlers-set!
	 wmii:client=? wmii:event-loop
	 wmii:color->string wmii:string->color
	 wmii:global-settings wmii:global-settings-set!
	 wmii:tag wmii:tag-settings wmii:tag-settings-set!
	 wmii:change-state wmii:kill wmii:navigate-to wmii:send-to
	 wmii:goto-tag wmii:client-tags wmii:client-tags-set!
	 wmii:tags wmii:clients wmii:write-tab wmii:destroy-tab
	 wmii:tabs wmii:quit wmii:exec))

(use 9p-client)

;; Workaround for Chicken < 3.0.1
(define (printf . args) (display (apply sprintf args)))

(define *wmii:connection* #f)

(define (camelcased->dasherized str)
  (list->string
   (reverse
    (string-fold
     (lambda (c l)
       (if (char-upper-case? c)
	   (if (not (null? l))
	       (cons (char-downcase c) (cons #\- l))
	       (cons (char-downcase c) l))
	   (cons c l))) '() str))))

(define (dasherized->camelcased str . rest)
  (let-optionals rest ((initial-caps #f))
    (let ((pieces (string-split str "-")))
      (if initial-caps
	  (apply string-append (map string-titlecase pieces))
	  (apply string-append (car pieces) (map string-titlecase (cdr pieces)))))))

(define (wmii:connect . rest)
  (let-optionals rest ((inport #f)
		       (outport #f))
   (set! *wmii:connection*
	 (if (and inport outport)
	     (9p:client-connect inport outport)
	     (call-with-values
		 (lambda ()
		   (use unix-sockets)
		   (unix-connect (sprintf "/tmp/ns.~A.~A/wmii"
					  (getenv "USER") (getenv "DISPLAY"))))
	       9p:client-connect)))
   *wmii:connection*))

(define (wmii:read file)
  (9p:with-input-from-file *wmii:connection* file read-string))

(define (wmii:read-lines file)
  (9p:with-input-from-file *wmii:connection* file read-lines))

(define (wmii:write file data)
  (9p:with-output-to-file *wmii:connection* file (lambda () (display data))))

(define (wmii:directory path)
  (9p:directory *wmii:connection* path))

(define (wmii:alist->rules alist)
  (apply string-append
   (map
    (lambda (rule)
      (sprintf "/~A/ -> ~A\n"
	       (car rule)
	       (if (pair? (cdr rule))
		   (string-join (map ->string (cdr rule)) "+")
		   (->string (cdr rule)))))
    alist)))

(define (wmii:rules->alist rules)
  (reverse
    (map
     (lambda (rule)
       (let ((result (string-match "^/([^/]+)/\\s*->\\s*(.*)$" rule)))
         (cons (second result) (string-split (third result) "+"))))
     rules)))

(define (wmii:colrules)
  (map string->number (wmii:rules->alist (wmii:read-lines "/colrules"))))

(define (wmii:colrules-set! rules)
  (wmii:write "/colrules" (wmii:alist->rules rules)))

(define (wmii:tagrules)
  (wmii:rules->alist (wmii:read-lines "/tagrules")))

(define (wmii:tagrules-set! rules)
  (wmii:write "/tagrules" (wmii:alist->rules rules)))

(define (wmii:key-code->string key-code)
  (string-join key-code "-"))

(define (wmii:string->key-code str)
  (string-split str "-"))

(define (wmii:grabbed-keys)
  (map wmii:string->key-code (wmii:read-lines "/keys")))

(define (wmii:grabbed-keys-set! keys)
  (9p:with-output-to-file *wmii:connection* "/keys"
    (lambda ()
      (for-each
       (lambda (key)
	 (printf "~a\n" (wmii:key-code->string key)))
       keys))))

(define *wmii:event-handlers* (list))

(define (wmii:event-handlers)
  *wmii:event-handlers*)

;; Option for not being smart about grabbed keys?
(define (wmii:event-handlers-set! handlers . rest)
  (let-optionals rest ((grab-keys #t))  
    (if grab-keys
	(let loop ((handlers handlers)
		   (keys '()))
	  (if (null? handlers)
	      (wmii:grabbed-keys-set! keys)
	      (match (car handlers)
		     ((('key . key-code) . _) (loop (cdr handlers) (cons key-code keys)))
		     (_ (loop (cdr handlers) keys)))))))
  (set! *wmii:event-handlers* handlers))

(define (wmii:parse-event line)
  (let* ((parts (string-split line))
	 (type (string->symbol (camelcased->dasherized (car parts)))))
    (case type
      ((key)
       (cons type (wmii:string->key-code (second parts))))
      ((urgent-tag not-urgent-tag)
       (list type (third parts) (string=? (second parts) "Client")))
      ((urgent not-urgent)
       (list type (second parts) (string=? (third parts) "Client")))
      ((client-mouse-down client-click)
       (list type (second parts) (string->number (third parts))))
      (else
       (cons type (cdr parts))))))

(define wmii:client=? string=?)

(define (wmii:handle-event event)
  (let ((handler (alist-ref event
			    *wmii:event-handlers*
			    (lambda (template event)
			      (if (pair? template)
				  (equal? template event)
				  (eq? template (car event)))))))
    (if handler
	(apply handler event))))

(define (wmii:event-loop . rest)
  (let-optionals rest ((kill-others #t))
    (if kill-others (wmii:write "/event" "Start wmiirc")) ;; Kill off any running wmiirc
    (9p:with-input-from-file *wmii:connection* "/event"
      (lambda ()
	(let loop ()
	  (let ((event (wmii:parse-event (read-line))))
	    (unless (equal? event '(start "wmiirc")) ;; Otherwise, return
		    (condition-case (wmii:handle-event event)
				    (exn (9p-server-error)
					 (fprintf (current-error-port)
						  "Server error: ~A\n"
						  ((condition-property-accessor 'exn 'message) exn))
					 (loop)))
		    (loop))))))))

(define (wmii:color->string color)
  (string-append "#" (string-pad (number->string color 16) 6 #\0)))

(define (wmii:string->color str)
  (string->number (string-drop str 1) 16))

(define (wmii:settings->alist settings)
 (map (lambda (line)
	(let* ((contents (string-split line))
	       (setting (cons (string->symbol (car contents)) (cdr contents))))
	  (case (car setting)
	    ((focuscolors normcolors)
	     (cons (car setting) (map wmii:string->color (cdr setting))))
	    (else
	     (if (null? (cddr setting))
		 (cons (car setting) (cadr setting))
		 setting)))))
      (if (pair? settings)
	  settings
	  (string-split settings "\n"))))

(define (wmii:alist->settings alist)
  (string-join
   (map (lambda (setting)
	  (case (car setting)
	    ((focuscolors normcolors) (sprintf "~A ~A\n" (car setting) (string-join (map wmii:color->string (cdr setting)))))
	    (else (if (pair? (cdr setting))
		      (sprintf "~A ~A\n" (car setting) (string-join (cdr setting)))
		      (sprintf "~A ~A\n" (car setting) (cdr setting))))))
	alist)))

(define (wmii:global-settings)
  (wmii:settings->alist (wmii:read "/ctl")))

(define (wmii:global-settings-set! alist)
  (wmii:write "/ctl" (wmii:alist->settings alist)))

(define (wmii:tag . rest)
  (let-optionals rest ((tag "sel"))
    (car (wmii:read-lines (sprintf "/tag/~A/ctl" tag)))))

(define (wmii:tag-settings . rest)
  (let-optionals rest ((tag "sel"))
    (wmii:settings->alist
     (cdr (wmii:read-lines (sprintf "/tag/~A/ctl" tag))))))

(define (wmii:tag-settings-set! alist . rest)
  (let-optionals rest ((tag "sel"))
    (wmii:write (sprintf "/tag/~A/ctl" tag) (wmii:alist->settings alist))))

(define (wmii:state-transition->string b)
  (case b
    ((#t) "on")
    ((#f) "off")
    ((toggle) "toggle")
    (else (error (sprintf "Unknown state transition type ~S" b)))))

(define (wmii:string->state-transition s)
  (cond
   ((string=? s "on") #t)
   ((string=? s "off") #f)
   ((string=? s "toggle") 'toggle)
   (else (error (sprintf "Unknown state transition type ~S" s)))))

;; Unfortunately, there's no wmii:client-settings because this information is
;; not exported by wmii.
(define (wmii:change-state state value . rest)
  (let-optionals rest ((client "sel"))
    (wmii:write (sprintf "/client/~A/ctl" client)
		(sprintf "~a ~a" state (wmii:state-transition->string value)))))

(define (wmii:kill . rest)
  (let-optionals rest ((client "sel"))
    (wmii:write (sprintf "/client/~A/ctl" client) "kill")))

(define (wmii:navigate-to where . rest)
  (let-optionals rest ((tag "sel"))
   (wmii:write (sprintf "/tag/~A/ctl" tag) (sprintf "select ~A" where))))

(define (wmii:send-to direction . rest)
  (let-optionals rest ((client "sel")
		       (tag "sel"))
   (wmii:write (sprintf "/tag/~A/ctl" tag)
	       (sprintf "send ~A ~A" client direction))))

(define (wmii:goto-tag tag)
  (wmii:write "/ctl" (sprintf "view ~A" tag)))

(define (wmii:client-tags . rest)
  (let-optionals rest ((client "sel"))
   (string-split (wmii:read (sprintf "/client/~A/tags" client)) "+")))

(define (wmii:client-tags-set! tags . rest)
  (let-optionals rest ((client "sel"))
   (wmii:write (sprintf "/client/~A/tags" client) (string-join tags "+"))))

(define (wmii:tags)
  (delete "sel" (9p:directory *wmii:connection* "/tag")))

(define (wmii:clients . tags)
  (let ((clients (delete "sel" (9p:directory *wmii:connection* "/client"))))
    (if (null? tags)
        clients
        (filter! (lambda (c)
                   (any (lambda (x) (member x tags string=?))
                        (wmii:client-tags c)))
                 clients))))

(define (wmii:write-tab bar tab contents . rest)
  (let-optionals rest ((colors #f))
   (wmii:write (sprintf "~A/~A" bar tab)
	       (if colors
		   (sprintf "~A ~A" (string-join (map wmii:color->string colors)) contents)
		   contents))))

(define (wmii:destroy-tab bar tab)
  (9p:delete-file *wmii:connection* (sprintf "~A/~A" bar tab)))

(define (wmii:tabs bar)
  (wmii:directory bar))

(define (wmii:quit)
  (wmii:write "/ctl" "quit"))

(define (wmii:exec cmdline)
  (wmii:write "/ctl" (sprintf "exec ~A" cmdline)))
